/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * <BR><BR>
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * <BR><BR>
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * <BR><BR>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * <BR><BR>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * <BR><BR>
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */
package org.greenstone.gatherer.gui;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import org.greenstone.gatherer.Configuration;
import org.greenstone.gatherer.Dictionary;
import org.greenstone.gatherer.Gatherer;
import org.greenstone.gatherer.cdm.Argument;
import org.greenstone.gatherer.collection.BuildOptions;
import org.greenstone.gatherer.collection.Collection;
import org.greenstone.gatherer.collection.CollectionManager;
import org.greenstone.gatherer.util.AppendLineOnlyFileDocument;
import org.greenstone.gatherer.util.AppendLineOnlyFileDocumentOwner;
import org.greenstone.gatherer.util.Utility;

/** This class serves as the data holder for all subclasses of option panes, such as Import options or All options. It also contains methods for creating each of the option lines as they would appear in the subpane. Futhermore it has a method for considering all the arguments and generating a <strong>String[]</strong> to allow you to pass them to the <strong>GShell</strong>.
 * @author John Thompson, Greenstone Digital Library, University of Waikato
 * @version 2.2
 */
public class OptionsPane
    extends JPanel 
    implements AppendLineOnlyFileDocumentOwner, MouseListener {

    static final public char SUCCESSFUL = 's';
    static final public char UNSUCCESSFUL = 'u';
    static final public char CANCELLED = 'c';
    static final public char UNKNOWN = 'x';

    static private int BUILD = 0;
    static private int IMPORT = 1;
    static private int MINIMUM_ROWS = 11;
    static private Dimension LABEL_SIZE = new Dimension(180, 25);
    static private Dimension ROW_SIZE = new Dimension(610, 30);
    static private Dimension SPINNER_SIZE = new Dimension(100, 25);
    static private String DESCRIPTION_SEP = " + ";

    /** All process messages are written to this log text area. */
    public JTextArea log_textarea = null;

    private ArrayList current_controls;

    /** The <strong>BuildOptions</strong> data object contains all the option settings we wish to persist between Gatherer sessions (and thus is stored in <strong>Collection</strong>). */
    private BuildOptions build_options = null;

    private FileEntry file_entry = null;

    /** the log pane - we only create it once now, not each time */
    private JPanel log_pane = null;
    /** the list of previous log messages */
    private JList log_list = null;
    private Vector writing_documents;


    /** The default constructor creates the few session length options, but either retrieves the rest from the current collection, or creates a default set of options. */
    public OptionsPane(BuildOptions build_options) {
	this.build_options = build_options;
	this.current_controls = new ArrayList();
	this.writing_documents = new Vector();

	// Have to do this here, not in display, as the message log view may not have been displayed yet.
	log_textarea = new JTextArea();
	log_textarea.setEditable(false);
    }

    /** This method creates the panel with all the build only options on it.
     * @param pane a JPanel which already has previous arguments available on it, to allow for lower moes to concatenate all of the arguments together
     * @return a JPanel which can be used to display all the build only options
     * @see org.greenstone.gatherer.Configuration#EXPERT_MODE
     * @see org.greenstone.gatherer.Configuration#getColor
     * @see org.greenstone.gatherer.Configuration#getMode
     * @see org.greenstone.gatherer.Gatherer#config
     * @see org.greenstone.gatherer.cdm.Argument
     * @see org.greenstone.gatherer.collection.BuildOptions#getBuildArgument
     * @see org.greenstone.gatherer.collection.BuildOptions#getBuildArgumentCount
     * @see org.greenstone.gatherer.collection.BuildOptions#getBuildValue
     * @see org.greenstone.gatherer.collection.BuildOptions#getBuildValueEnabled
     * @see org.greenstone.gatherer.gui.OptionsPane.ArgumentControl
     */
    public JPanel buildBuild(JPanel pane) {
	// Reset the arguments
	if(pane == null) {
	    current_controls.clear();
	}
	ArrayList build_arguments = new ArrayList();
	int current_mode = Gatherer.config.getMode();
	int total_build_argument_count = build_options.getBuildArgumentCount();       
	for(int i = 0; i < total_build_argument_count; i++) {
	    // Retrieve the argument so we know how to format the control.
	    Argument argument = build_options.getBuildArgument(i);
	    if(!argument.isHiddenGLI() && argument.getModeLevel() <= current_mode) {
		// Now attempt to retrieve any existing value for this argument.
		boolean enabled = build_options.getBuildValueEnabled(argument.getName());
		String value = build_options.getBuildValue(argument.getName());
		ArgumentControl argument_control = new ArgumentControl(BUILD, argument, enabled, value);
		build_arguments.add(argument_control);
	    }
	}
	current_controls.addAll(build_arguments);
	// Now that we know how many arguments there are we can build the pane to view them on. Modes lower than EXPERT can provide a previous pane on which to add the arguments.
	if(pane == null || current_mode >= Configuration.EXPERT_MODE) {
	    pane = new JPanel();
	    pane.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
	    pane.setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
	    int argument_count = build_arguments.size();
	    // If in any of the higher detail modes, and assuming we don't want super phat argument controls, we better ensure there is a minimum number or lines in the grid layout
	    if(current_mode >= Configuration.EXPERT_MODE) {
		if(argument_count < MINIMUM_ROWS) {
		    argument_count = MINIMUM_ROWS;
		}
		pane.setLayout(new GridLayout(argument_count, 1, 5, 5));
	    }
	    // Otherwise we're just going to throw them on one after another and chuck it in a scroll pane anyway
	    else {
		pane.setLayout(new BoxLayout(pane, BoxLayout.Y_AXIS));
	    }
	}
	for(int j = 0; j < build_arguments.size(); j++) {
	    pane.add((JComponent)build_arguments.get(j));
	}
	pane.addMouseListener(this);
	build_arguments = null;
	return pane;
    }

    /** This method creates the panel with all the import only options on it.
     * @param pane a JPanel which already has previous arguments available on it, to allow for lower moes to concatenate all of the arguments together
     * @return a JPanel which can be used to display all the build only options
     * @see org.greenstone.gatherer.Configuration#EXPERT_MODE
     * @see org.greenstone.gatherer.Configuration#getColor
     * @see org.greenstone.gatherer.Configuration#getMode
     * @see org.greenstone.gatherer.Gatherer#config
     * @see org.greenstone.gatherer.cdm.Argument
     * @see org.greenstone.gatherer.collection.BuildOptions#getImportArgument
     * @see org.greenstone.gatherer.collection.BuildOptions#getImportArgumentCount
     * @see org.greenstone.gatherer.collection.BuildOptions#getImportValue
     * @see org.greenstone.gatherer.collection.BuildOptions#getImportValueEnabled
     * @see org.greenstone.gatherer.gui.OptionsPane.ArgumentControl
     */
    public JPanel buildImport(JPanel pane) {
	// Reset the arguments
	if(pane == null) {
	    current_controls.clear();
	}
	ArrayList import_arguments = new ArrayList();
	int current_mode = Gatherer.config.getMode();
	int total_import_argument_count = build_options.getImportArgumentCount();       
	for(int i = 0; i < total_import_argument_count; i++) {
	    // Retrieve the argument so we know how to format the control.
	    Argument argument = build_options.getImportArgument(i);
	    if(!argument.isHiddenGLI() && argument.getModeLevel() <= current_mode) {
		// Now attempt to retrieve any existing value for this argument.
		boolean enabled = build_options.getImportValueEnabled(argument.getName());
		String value = build_options.getImportValue(argument.getName());
		ArgumentControl argument_control = new ArgumentControl(IMPORT, argument, enabled, value);
		import_arguments.add(argument_control);
	    }
	}
	current_controls.addAll(import_arguments);
	// Now that we know how many arguments there are we can build the pane to view them on. Modes lower than EXPERT can provide a previous pane on which to add the arguments.
	if(pane == null || current_mode >= Configuration.EXPERT_MODE) {
	    pane = new JPanel();
	    pane.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
	    pane.setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
	    int argument_count = import_arguments.size();
	    // If in any of the higher detail modes, and assuming we don't want super phat argument controls, we better ensure there is a minimum number or lines in the grid layout
	    if(current_mode >= Configuration.EXPERT_MODE) {
		if(argument_count < MINIMUM_ROWS) {
		    argument_count = MINIMUM_ROWS;
		}
		pane.setLayout(new GridLayout(argument_count, 1, 5, 5));
	    }
	    // Otherwise we're just going to throw them on one after another and chuck it in a scroll pane anyway
	    else {
		pane.setLayout(new BoxLayout(pane, BoxLayout.Y_AXIS));
	    }
	}
	for(int j = 0; j < import_arguments.size(); j++) {
	    pane.add((JComponent)import_arguments.get(j));
	}
	pane.addMouseListener(this);
	import_arguments = null;
	return pane;
    }

    /** This method is used to build a panel based on the message log, which is nothing like any of the other panels.
     * @return A <strong>JPanel</strong> containing a scrollable text area which represents the shell process message log.
     */
    public JPanel buildLog() {
	// we now save the log pane
	if (log_pane == null) {
	    log_pane = new JPanel(new BorderLayout());

	    // Build a list of the log files available, ordering by last modified. Log files are like build_log.date.txt
	    DefaultListModel contents = new DefaultListModel();
	    File log_directory = new File(Gatherer.c_man.getCollectionLog());
	    File children[] = log_directory.listFiles();
	    for(int i = 0; children != null && i < children.length; i++) {
		String filename = children[i].getName();
		if(filename.startsWith("build_log.") && filename.endsWith(".txt") ) {
		    String datestamp = filename.substring(filename.indexOf(".") + 1, filename.lastIndexOf(".")).toLowerCase();
		    if(datestamp.indexOf("s") == -1 && datestamp.indexOf("u") == -1 && datestamp.indexOf("c") == -1 && datestamp.indexOf("x") == -1) {
			FileEntry entry = new FileEntry(children[i].getName(), children[i].getAbsolutePath());
			// We are about to insert it. But where.
			boolean found = false;
			for(int j = 0; !found && j < contents.size(); j++) {
			    FileEntry sibling = (FileEntry) contents.getElementAt(j);
			    int order = entry.compareTo(sibling);
			    if(order > 0) {
				contents.insertElementAt(entry, j);
				found = true;
			    }
			}
			if(!found) {
			    contents.addElement(entry);
			}
		    }
		}
	    }

	    log_list = new JList(contents);
	    log_list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	    log_list.setLayoutOrientation(JList.VERTICAL);
	    log_list.setPreferredSize(new Dimension(600, 100));
	    log_list.setVisibleRowCount(3);
	    log_list.addListSelectionListener(new LogListListener());

	    JLabel log_history_label = new JLabel();
	    Dictionary.registerText(log_history_label, "OptionsPane.LogHistory");
	    JPanel log_history_pane = new JPanel();
	    log_history_pane.setPreferredSize(new Dimension(600, 100));
	    log_history_pane.setLayout(new BorderLayout());
	    log_history_pane.add(log_history_label, BorderLayout.NORTH);
	    log_history_pane.add(new JScrollPane(log_list), BorderLayout.CENTER);
	    
	    log_pane.add(new JScrollPane(log_textarea), BorderLayout.CENTER);
	    log_pane.add(log_history_pane, BorderLayout.SOUTH);
	}
	return log_pane;
    }

    public AppendLineOnlyFileDocument createNewLogDocument() {
	long time = System.currentTimeMillis();
	StringBuffer name = new StringBuffer();
	name.append("build_log.");
	name.append(time);
	name.append(".txt");
	// just in case there is no log directory
	File file = new File(Gatherer.c_man.getCollectionLog() + name.toString());
	File parent_file = file.getParentFile();
	parent_file.mkdirs();
	parent_file = null;
	// create the file entry and add it to the list at pos 0 - it will always be the newest one created
	file_entry = new FileEntry(name.toString(), file.getAbsolutePath());
	((DefaultListModel)log_list.getModel()).add(0, file_entry);
	log_list.setSelectedIndex(0);
	// Finally retrieve and return the document associated with this file entry
	return file_entry.getDocument();
    }


    /** Attempts to discover the latest document count.
     * @return An <strong>int</strong> detailing the number of documents in this collection.
     */
    public int getDocumentCount() {
	if(Gatherer.c_man.ready()) {
	    int count = Gatherer.c_man.getCollection().getDocumentCount();
	    if(count != 0) {
		return count;
	    }
	}
	return 1;
    }

    /** Called by our magic log documents after they have finished writing themselves to file, whereapon it is no longer necessary to hold a reference to them. */
    public void remove(AppendLineOnlyFileDocument document) {
	writing_documents.remove(document);
    }

    public void resetFileEntry() {
	if(file_entry != null) {
	    file_entry.reset();
	}
    }

    /** Given a panel containing ArgumentControls, update the values associated with them. */
    public void update(JPanel panel) {
	if(panel == log_pane) {
	    return;
	}

	for(int i = 0; i < panel.getComponentCount(); i++) {
	    Component component = panel.getComponent(i);
	    if(component instanceof ArgumentControl) {
		((ArgumentControl)component).update();
	    }
	}
    }

    /** Implementation side-effect
     * @param e a MouseEvent
     */
    public void mouseClicked(MouseEvent e) {}

    /** Implementation side-effect
     * @param e a MouseEvent
     */
    public void mouseEntered(MouseEvent e) {}

    /** Implemented to ensure that, by the time the mouse pointer leaves the current build options screen, ang changes to the value the JSpinners have been commited
     * @param e a MouseEvent
     */
    public void mouseExited(MouseEvent e) {
	// Loop through the controls, and if the current control is a JSpinner, commit its current editing
	for(int i = 0; i < current_controls.size(); i++) {
	    ArgumentControl control = (ArgumentControl) current_controls.get(i);
	    JComponent value_control = control.getValueControl();
	    if(value_control instanceof JSpinner) {
		try {
		    ((JSpinner)value_control).commitEdit();
		}
		catch(Exception exception) {
		    Gatherer.println("Exception in OptionsPane.mouseExited() - unexpected");
		    Gatherer.printStackTrace(exception);
		}
	    }
	    value_control = null;
	    control = null;
	}
    }

    /** Implementation side-effect
     * @param e a MouseEvent
     */
    public void mousePressed(MouseEvent e) {}

    /** Implementation side-effect
     * @param e a MouseEvent
     */
    public void mouseReleased(MouseEvent e) {}

    private class ArgumentControl
	extends JPanel {
	private Argument argument;
	private int type;
	private JComponent value_control;
	private JCheckBox enabled;
	public ArgumentControl(int type, Argument argument, boolean enable, String value) {
	    super();
	    this.argument = argument;
	    this.type = type;
	    String tooltip = Utility.formatHTMLWidth("<html>" + argument.getDescription() + "</html>", 60);
	    // Because of the dynamic order of component creation/connection/layout, we can't really follow that pattern here.
	    setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
	    setBorder(BorderFactory.createEmptyBorder(2,0,2,0));
	    setLayout(new BorderLayout());
	    setPreferredSize(ROW_SIZE);
	    setMaximumSize(ROW_SIZE);

	    // Try to determine what the value_controls value should be.
	    if(value == null) {
		value = argument.getDefaultValue();
	    }
	    // Create a correct value control based on the argument provided.
	    switch(argument.getType()) {
	    case Argument.ENUM:
	    case Argument.METADATUM:
		JComboBox combobox = new JComboBox();
		combobox.setEnabled(enable);
		combobox.setToolTipText(tooltip);
		combobox.setBackground(enable ? Color.white : Color.lightGray);

		// Build an option model, wrapping each entry of the list table
		if (argument.getType() == Argument.ENUM) {
		    HashMap arg_list = argument.getOptions();
		    Iterator it = arg_list.keySet().iterator();
		    while (it.hasNext()) {
			combobox.addItem((String) it.next());
		    }

		    // Connect this so if a value is selected the tooltip updates accordingly
		    combobox.addActionListener(new ToolTipUpdater(arg_list));
		}
		if (argument.getType() == Argument.METADATUM) {
		    Vector meta_list = Gatherer.c_man.getCollection().msm.getAssignedElements();
		    for (int i = 0; i < meta_list.size(); i++) {
			combobox.addItem(meta_list.get(i));
		    }
		}

		if (value != null) {
		    // Set the selected string
		    for (int i = 0; i < combobox.getItemCount(); i++) {
			if (combobox.getItemAt(i).toString().equals(value)) {
			    combobox.setSelectedIndex(i);
			}
		    }
		}
		// Layout
		add(combobox, BorderLayout.CENTER);
		// And remember
		value_control = combobox;
		break;
	    case Argument.FLAG:
		// Only need the check box.
		value_control = null;
		break;
	    case Argument.INTEGER:
		// Build a spinner
		int initial_value=0;
		if (value != null) {
		    try {
			initial_value = Integer.parseInt(value);
		    } catch (Exception exception) {
		    }
		}
		if (initial_value < argument.getMinimum()) {
		    initial_value = argument.getMinimum();
		} else if (initial_value > argument.getMaximum()) {
		    initial_value = argument.getMaximum();
		}
		JSpinner spinner = new JSpinner(new SpinnerNumberModel(initial_value, argument.getMinimum(), argument.getMaximum(), 1));		
		spinner.setEnabled(enable);
		spinner.setPreferredSize(SPINNER_SIZE);
		spinner.setToolTipText(tooltip);
		// Set enabled
		JComponent c = spinner.getEditor();
		if ( c instanceof JSpinner.DefaultEditor ) {
		    JSpinner.DefaultEditor editor = (JSpinner.DefaultEditor) c;
		    JFormattedTextField field = editor.getTextField();
		    field.setEditable(enable);
		    if(enable) {
			field.setBackground(Color.white);
		    }
		    else {
			field.setBackground(Color.lightGray);
		    }
		}
		// Layout
		add(new JLabel(), BorderLayout.CENTER);
		add(spinner, BorderLayout.EAST);
		// And remember it
		value_control = spinner;
		break;
	    case Argument.STRING:
		// Use a standard text field
		JTextField textfield = new JTextField();
		textfield.setEnabled(enable);
		textfield.setToolTipText(tooltip);
		// Set enabled
		if(enable) {
		    textfield.setBackground(Color.white);
		}
		else {
		    textfield.setBackground(Color.lightGray);
		}
		// If there was an original value, set it.
		if(value != null) {
		    textfield.setText(value);
		}
		// Layout
		add(textfield, BorderLayout.CENTER);
		// And remember it
		value_control = textfield;
		break;
	    }

	    // If the argument is required, then you don't get a choice of whether it is enabled.
	    if(argument.isRequired()) {
		JLabel label = new JLabel(argument.getName());
		label.setOpaque(false);
		label.setPreferredSize(LABEL_SIZE);
		label.setToolTipText(tooltip);
		add(label, BorderLayout.WEST);
	    }
	    else {
		enabled = new JCheckBox(argument.getName(), enable);
		enabled.setOpaque(false);
		enabled.setPreferredSize(LABEL_SIZE);
		enabled.setToolTipText(tooltip);
		// Connect
		enabled.addActionListener(new EnabledListener(value_control));
		// Layout
		add(enabled, BorderLayout.WEST);
	    }
	}

	/** Retrieve the control used for storing values.
	 * @return JComponent
	 */
	public JComponent getValueControl() {
	    return value_control;
	}

	/** Method to ensure that a certain value is selected, if it exists within that combobox to begin with.
	 * @param combobox the JComboBox whose selection we are trying to preset
	 * @param target The desired value of the selection as a String
	 * @return true if the item was found and selected, false otherwise
	 */
	public boolean selectValue(JComboBox combobox, String target) {
	    for(int i = 0; i < combobox.getItemCount(); i++) {
		if(combobox.getItemAt(i).toString().equals(target)) {
		    combobox.setSelectedIndex(i);
		    return true;
		}
	    }
	    return false;
	}

	/** Update the values stored in the collection so as to rememebr the current state of this argument. */
	public void update() {
	    String name = argument.getName();
	    boolean enable = true;
	    if(enabled != null) {
		enable = enabled.isSelected();
	    }
	    String value = null;
	    if(value_control == null) {
		// Flag value, nothing to do.
	    }
	    else if(value_control instanceof JTextField) {
		value = ((JTextField)value_control).getText();
	    }
	    else if(value_control instanceof JSpinner) {
		value = ((JSpinner)value_control).getValue().toString();
	    }
	    else if(value_control instanceof JComboBox) {
		value = (((JComboBox)value_control).getSelectedItem()).toString();
	    }
	    // If this argument was a flag, but is now disabled, remove from the build options altogether
	    if(!enable && value == null) {
		if(type == BUILD) {
		    build_options.removeBuildValue(name);
		}
		else {
		    build_options.removeImportValue(name);
		}
	    }
	    // Otherwise update the argument value
	    else {
		if(type == BUILD) {
		    build_options.setBuildValue(name, enable, value);
		}
		else {
		    build_options.setImportValue(name, enable, value);
		}
	    }
	}
    }

    /** Listens for actions apon the enable checkbox, and if detected enables or diables control appropriately. */
    private class EnabledListener
	implements ActionListener {
	/** An editor component, such as a JComboBox or JTextField, that might have its enabled state changed by this listener. */
	private JComponent target = null;
	/** Constructor. */
	public EnabledListener(JComponent target) {
	    this.target = target;
	}
	/** Any implementation of ActionListener must include this method so that we can be informed when an action has been performed on or registered check box, prompting us to change the state of the other controls as per the users request.
	 * @param event An <strong>ActionEvent</strong> containing information about the click.
	 */
	public void actionPerformed(ActionEvent event) {
	    JCheckBox source = (JCheckBox)event.getSource();
	    if(target != null) {
		if(source.isSelected()) {
		    target.setBackground(Color.white);
		    target.setEnabled(true);
		}
		else {
		    target.setBackground(Color.lightGray);
		    target.setEnabled(false);
		}
		// Special case of stupid JSpinners who don't let their backgrounds change properly.
		if(target instanceof JSpinner) {
		    JSpinner spinner = (JSpinner) target;
		    JComponent c = spinner.getEditor();
		    if ( c instanceof JSpinner.DefaultEditor ) {
			JSpinner.DefaultEditor editor = (JSpinner.DefaultEditor) c;
			JFormattedTextField field = editor.getTextField();
			field.setEditable(source.isSelected());
			if(source.isSelected()) {
			    field.setBackground(Color.white);
			}
			else {
			    field.setBackground(Color.lightGray);
			}
		    }
		}
	    }
	}
    }

    /** Holds a File which has a particular naming convention build_log.date.txt also keeps a Date corresponding to the date in its name*/
    private class FileEntry {

       private AppendLineOnlyFileDocument current_document;
       private Date date;
       private long last_modified;
       private String display;
       private String filename;
       private String filepath;
       
       public FileEntry(String filename, String filepath) {
	  this.date = null;
	  this.display = null;
	  this.filename = filename;
	  this.filepath = filepath;
	  this.last_modified = 0L;
       }

	/** returns 0 if the dates are the same, -ve number if the current FileEntry is earlier than the fe FileEntry ...*/
	public int compareTo(FileEntry file_entry) {
	    Date our_date = getDate();
	    Date other_date = file_entry.getDate();
	    return our_date.compareTo(other_date);
	}

	public Date getDate() {
	    if(date == null) {
		// Need to exclude first '.'
		int first_index = filename.indexOf(".") + 1;
		// Need to exclude the last '.'
		int last_index = filename.lastIndexOf(".");
		if(first_index > 0 && last_index > 0 && first_index < last_index) {
		    String date_string = filename.substring(first_index, last_index);
		    date = new Date(Long.parseLong(date_string));
		}
		else {
		    date = new Date(); // Current date
		}
	    }
	    return date;
	}

       public AppendLineOnlyFileDocument getDocument() {
	  if(current_document == null) {
	     current_document = new AppendLineOnlyFileDocument(filepath);
	  }
	  return current_document;
	}

	public void reset() {
	    display = null;
	}

	/** we only want the date out of the file name, not the whole path */
	public String toString() {
	    File file = new File(filename);
	    if(display == null) {
		last_modified = file.lastModified();
		StringBuffer d = new StringBuffer();
		Date date = getDate();
		d.append(date.toString());
		char success = UNKNOWN;
		File the_file = new File(filepath);
		if(the_file.exists()) {
		   try {
		      FileInputStream in = new FileInputStream(the_file);
		      success = (char) in.read();
		      in.close();
		      in = null;
		   }
		   catch(Exception error) {
		      ///ystem.err.println("Log '" + filepath + "' not found!");
		      ///atherer.printStackTrace(error);
		   }
		}
		the_file = null;
		switch (success) {
		case SUCCESSFUL:
		    d.append(Dictionary.get("OptionsPane.Successful"));
		    break;
		case UNSUCCESSFUL:
		    d.append(Dictionary.get("OptionsPane.Unsuccessful"));
		    break;
		case CANCELLED:
		    d.append(Dictionary.get("OptionsPane.Cancelled"));
		    break;
		default:
		    d.append(Dictionary.get("OptionsPane.Unknown"));
		}
		display = d.toString();
	    }
	    return display;
	}
    }

    /** a ListSelectionListener that triggers the load of a newly selected log */
    private class LogListListener implements ListSelectionListener {

	public void valueChanged(ListSelectionEvent e) {
	    if (!e.getValueIsAdjusting()) { // we get two events for one change in list selection - use the false one ( the second one)
		///ystem.err.println("Log change detected.");
		JList source  = (JList)e.getSource();
		file_entry = (FileEntry) source.getSelectedValue();
		// First we determine if the old log has been completely written to file
		Document document = log_textarea.getDocument();
		///ystem.err.println(" * current document: " + document);
		///ystem.err.println(" * new document:     " + file_entry.getDocument());
		// If we are dealing with the same document don't do anything.
		if(document != file_entry.getDocument()) {
		    if(document instanceof AppendLineOnlyFileDocument) {
			AppendLineOnlyFileDocument append_line_only_file_document = (AppendLineOnlyFileDocument) document;
			if(append_line_only_file_document.isStillWriting()) {
			    ///ystem.err.println("Current log is still active... finishing.");
			    writing_documents.add(append_line_only_file_document); // We have to maintain a reference until they are all done.
			    append_line_only_file_document.setOwner(OptionsPane.this);
			    append_line_only_file_document.setExit();
			}
			else {
			    ///ystem.err.println("Current log is complete. Nothing to do.");
			}
		    }
		    // Load the new log
		    log_textarea.setDocument(file_entry.getDocument());
		}
	    }
	}
    }

    /** Listener that sets the tooltip associated to a combobox to the tooltip relevant to the selected item. */
    private class ToolTipUpdater
	implements ActionListener {
	private HashMap arg_list;
	public ToolTipUpdater(HashMap arg_list) {
	    this.arg_list = arg_list;
	}
	/** Any implementation of an ActionListener must include this method so that we can be informed when the selection in a combobox has changed and update the tooltip accordingly.
	 * @param event An <strong>ActionEvent</strong> containing information about the action that fired this call.
	 */
	public void actionPerformed(ActionEvent event) {
	    JComboBox source = (JComboBox)event.getSource();
	    String key = (String) source.getSelectedItem();
	    if(arg_list != null) {
		String description = (String) arg_list.get(key);
		if(description != null) {
		    description = Utility.formatHTMLWidth(DESCRIPTION_SEP + description, 60);
		    String original = source.getToolTipText();
		    if(original == null) {
			original = "";
		    }
		    // Remove any existing extra description.
		    if(original.indexOf(DESCRIPTION_SEP) != -1) {
			original = original.substring(0, original.indexOf(DESCRIPTION_SEP));
		    }
		    source.setToolTipText(original + description);
		}
	    }
	}
    }
}
